/*********** Generating random Tanner-graphs with large girth *************
 * 
 * Generate a bipartite graph with n variable nodes of degree deg_n and 
 * m check nodes of degree deg_m
 * k : 	Eliminate all loops of girth 2k + 2.
 * jseed : seed for the random number gen. 
 * 	(Set jseed = 0 for use the current time).
 * File Identifier : string to append to the name of the output file 
 * 	(eg. v1, v2 etc.)
 * 
 * Output format : 
 * 	First row : n
 * 	Second row : m
 * 	Followed by edge list.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include<string.h>
#include "svdlib.h"


int n;
int m;
int kgir;

double **E ;
//int** A1;
double **bipConst;

double **W;

int deg1, deg2 ;
double** Lam;

int *rdeg;
int *cdeg;
FILE *fout;


//Main Function:
int *d2w;
double lam;

void alloc_mem(void);
void free_mem(void);

double absolute(double x);
double Weight(int et);
void mulByW4(SMat St, double **Q, double **Q1);
void mulByW(SMat St, double **Q, double **Z);

void initW(int flag);

int addedge(void); 
void addeye(double** SS, int n1,double **out,int flag);

void sparsemul(SMat A, double** B,int n1, int n2, int n3, double** C, int flag);
void matrixmul(double** A, double** B, int n1, int n2, int n3, double** C,int flag, double alpha, double beta);


void output(char* outfile, int t);

int pick(double *probs,int nmax);
//General:
double factorial(int a, int b);
double rannyu(void);
void setrn(int iseed[]) ;
double** callocArray(int n1, int n2);
double** mallocArray(int n1, int n2);
void freeArray(double **A, int n1);

double max1(double a, double b);
int sign1(int x);


int main(int argc, char** argv)
{
	int i,j;
	int t;
	double qsum;
	int iseed[4];
    int jseed;
	char* outfile ;
	int e;
fout = fopen("output","a");

	if (argc==8)
    {
        n = atoi(argv[1]);
		deg1 = atoi(argv[2]);
		m = atoi(argv[3]);
		deg2 = atoi(argv[4]);
		kgir = atoi(argv[5]);
		jseed = atoi(argv[6]);
		outfile = argv[7] ;
    }
    else
    {
        printf("\n# Usage:  %s  <n>  <deg_n>  <m>  <deg_m>  <k>   <jseed> <File Identifier>\n\n",argv[0]);
	printf("Generate a bipartite graph with n variable nodes of degree deg_n and m check nodes of degree deg_m\n");
	printf("k : Eliminate all loops of girth 2k + 2.\n");
	printf("jseed : seed for the random number gen. (Set jseed = 0 for use the current time).\n");
	printf("File Identifier : string to append to the name of the output file\n\n");
        return;
    }



/****************************************/
// Seed etc.
 	if(jseed == 0)
		jseed = time(NULL);
	printf("%d\n",jseed	) ;
	srand(jseed) ;
	for(i=0; i<4; ++i)
		iseed[i] = rand() % 4096;
	
	setrn(iseed);

/***************************************/




	int fail;
int iter, failtot=0;
int dummy;
for(iter = 0; iter<1; ++iter)
{
	fail = 1;
	alloc_mem();

/***************************************/
// Initialize degree distribution :
	e = 0;
	for(i=0; i<n; ++i)
	{
		rdeg[i] = deg1;
		e = e + rdeg[i];
	}
	

	for(i=0; i<m; ++i)
		cdeg[i] = deg2;
	
/***************************************/


// Main Loop :
	//printf("%d\n",e);
	double ratio = 1.99;
//	printf("%d\n",e);
	for(t=0; t<e; ++t)
	{
		
		
		d2w[0] = 0;
		for(i=1; i<=n; ++i)
			d2w[i] = d2w[i-1] + 2*(rdeg[i-1]+1);
		for(i=n+1; i<=m+n; ++i)
			d2w[i] = d2w[i-1] + 2*(cdeg[i-n-1]+1) ;

		
		lam = 1.0/((double)(e-t)) ;
		if( t < ratio*e )
		{	
			initW(1);
			Weight(d2w[m+n] );
		}
		else
		{
			initW(0) ;
		}
		
		output(outfile,t);
		//printf("%d \n",t) ;
		fail = addedge() ;
		if( fail == 0 )
		{
//			printf("Fail at t = %d\n",t);
			break;
		}	
		
	}
	//printf("\n");
/**************************************/


	failtot += fail ;

	printf("%d %d/%d %f\n",t,failtot,(iter+1),rannyu());
	output(outfile,t);
	free_mem();
}//End of iter loop







	fclose(fout);

	return(0);
}


double Weight(int et)
{
	int nnz = 0;


	int i,j,ind,ind1,ind2,count;

	int *dv;
	dv = calloc(et,sizeof(int));
	double **Q, **Q1;
	Q = (double**) callocArray(et,et);
	Q1 = (double**) callocArray(et,et);



	for(i=n; i<n+m; ++i)
		for(j=0; j<n; ++j)
			if( ((int) E[i][j] ) == 1 )
				nnz += 2*(rdeg[j]+1)*(cdeg[i-n]+1) ;


//	for(i=0; i<n; ++i)
//		for(j=n; j<n+m; ++j)
//			if( ((int) E[i][j] ) == 1 )
//				nnz += 2*rdeg[i]*cdeg[j-n] ;
	
	for(i=0; i<n; ++i)
		nnz += (rdeg[i]+1)*(rdeg[i]+1) ;
	for(i=0; i<m; ++i)
		nnz += (cdeg[i]+1)*(cdeg[i]+1) ;
	
	SMat St ;
	St = svdNewSMat( et,et,nnz) ;
//					for( ind1 = 0; ind1< cdeg[i-n]; ++ind1)

	count = 0;
	for(j = 0; j<n; ++j)
	{
		for(ind = 0; ind<rdeg[j]; ++ind)
			for(i = n; i<n+m; ++i)
				if( ( (int) E[i][j] ) == 1 )
					for( ind1 = 0; ind1< cdeg[i-n]; ++ind1)
					{
						St->rowind[count] = d2w[i] + cdeg[i-n] +1+ ind1 ;
						St->value[count] = -lam  ;
						dv[d2w[j]+ind] += 1 ;
						count += 1;
					}
					
		ind = rdeg[j];			
			for(i = n; i<n+m; ++i)
				if( ( (int) E[i][j] ) == 1 )
				{
					ind1 = cdeg[i-n]; 
//					ind1 = 0; 
					{
						St->rowind[count] = d2w[i] + cdeg[i-n] +1+ ind1 ;
						St->value[count] = 1 ;
						dv[d2w[j]+ind] += 1 ;
						count += 1;
					}
				}	
		

		for(ind = 0; ind<rdeg[j]+1; ++ind)
			for(ind1 = 0; ind1<rdeg[j]+1; ++ind1)
			{	

				St->rowind[count] = d2w[j] + ind1 ;
				St->value[count] = 1 ;
				if(ind1 == ind)
					St->value[count] = 0 ;
				
				dv[d2w[j]+rdeg[j]+1+ind] += 1;
				++count;
			}
				St->value[count] = 1;

	}		
	
	for(j = n; j<n+m; ++j)
	{
		for(ind = 0; ind<cdeg[j-n]; ++ind)
			for(i = 0; i<n; ++i)
				if( (int) E[i][j] == 1 )
					for( ind1 = 0; ind1< rdeg[i]; ++ind1)
					{
						St->rowind[count] = d2w[i] + rdeg[i] +1+ ind1 ;
						St->value[count] = -lam ;
						dv[d2w[j]+ind] += 1 ;
						count += 1;
					}

		ind = cdeg[j-n] ;			
			for(i = 0; i<n; ++i)
				if( (int) E[i][j] == 1 )
				{
					ind1 = rdeg[i];
					{
						St->rowind[count] = d2w[i] + rdeg[i] +1+ ind1 ;
						St->value[count] = 1;
						dv[d2w[j]+ind] += 1 ;
						count += 1;
					}
				}	
		
		for(ind = 0; ind<cdeg[j-n]+1; ++ind)
			for(ind1 = 0; ind1<cdeg[j-n]+1; ++ind1)
			{	

				St->rowind[count] = d2w[j] + ind1 ;
				St->value[count] = 1 ;
				if(ind1 == ind)
					St->value[count] = 0 ;
				
				dv[d2w[j]+cdeg[j-n]+ind] += 1;
				++count;
			}
				St->value[count] = 1;
	}		
		

	St->rows = et;
	St->cols = et;
	St->vals = count;


	St->pointr[0] = 0 ;

    for(i=1; i<= St->cols; ++i)
    	St->pointr[i] = St->pointr[i-1] + dv[i-1] ;
	

	addeye(Q,et,Q,0);
//	printf("%d %d\n",et,nnz-count);

	for(i = 0; i<kgir; ++i)
	{
		mulByW4(St,Q,Q1);
		addeye(Q,et,Q,0);
	}


	mulByW(St,Q,Q1) ;
	mulByW(St,Q1,Q) ;

//	addeye(Q,et,Q,0);
//	addeye(Q,et,Q,0);
	mulByW(St,Q,Q1) ;
//	addeye(Q1,et,Q1,0);

//	printf("ad ... \n");


	double fullsum = 0;
	double sum;

	double maxW = -1.0;
	double minW = 1e100;

	double **Wtemp;
	Wtemp = callocArray(m+n,m+n);


//	printf("%d %d %d\n",d2w[n],d2w[150],d2w[149]);
	for(i=0; i<n+m; ++i)
		for(j=0; j<n+m; ++j)
		{
			sum = 0;
//			printf("%d %d\n",i,j);
			for(ind1=d2w[i]; ind1<d2w[i+1];++ind1)
				for(ind2 = d2w[j] ; ind2<d2w[j+1] ; ++ind2)
				{
//					printf("%d %d\n",ind1,ind2);
					sum += Q1[ind1][ind2] ;
				}	
			
			if( W[i][j] > 1e-3 )
			{
				if( sum > maxW)
					maxW = sum ;
				if(sum < minW)
					minW = sum ;
			}		

			Wtemp[i][j] = Q1[d2w[i+1]-1][ d2w[j] +(d2w[j+1]-d2w[j])/2 - 1] ;			
//			Wtemp[i][j] = Q1[ d2w[i] +(d2w[i+1]-d2w[i])/2 - 1][d2w[j] + (d2w[j+1]-d2w[j])/2 -1] ;			
//			Wtemp[i][j] = sum ;
			fullsum += W[i][j] ;
		}


//	if( maxW < 30.0 )
	{
		for(i=0; i<n+m; ++i)
			for(j=0; j<n+m; ++j)
				W[i][j] *= exp(-Wtemp[i][j] );
	}			
/*	else
	{
		
		for(i=0; i<n+m; ++i)
			for(j=0; j<n+m; ++j)
			{
//				printf("%d %d\n",i,j);
				if( absolute(Wtemp[i][j] - minW) < 30.0 )
					W[i][j] *= exp(-Wtemp[i][j]+minW) ;
				else
					W[i][j] *= exp(-30.0) ;
			}		
					
	}
*/		
			

//	printf("bc ... \n");


	printf("%f %f ",maxW,minW);
	svdFreeSMat(St);
	freeArray(Q,et);
	freeArray(Q1,et);
	freeArray(Wtemp,m+n);
	free(dv);
	return(fullsum);
}


void mulByW4(SMat St, double **Q, double **Q1)
{
//	printf("b %d %d %d\n\n\n",et,nnz,count);
	mulByW(St,Q,Q1);
	mulByW(St,Q1,Q);
	mulByW(St,Q,Q1);
	mulByW(St,Q1,Q);
}



void mulByW(SMat St, double **Q, double **Z)
{
//Compute Z = (N+S_t)*Q 
//N_t*Q = A
//        C  



	int i,j,ind;

	double *A;
	double *B;

	B = (double*) calloc((int) St->rows, sizeof(double));
	A = (double*) calloc((int) St->rows, sizeof(double));
//	printf("%d\n",St->rows);
//	printf("%d %d\n",A-B,B);

//	printf("0\n");
	sparsemul(St,Q,St->rows,St->rows,St->rows,Z,0);
//	printf("a\n");
	for(i=n; i<m+n; ++i)
		for(ind = 0; ind < cdeg[i-n]; ++ind)
		{
			for(j=0; j<St->rows; ++j)
				A[j] += Q[d2w[i]+ind][j] ;

		}
	
	for(i=0; i<n; ++i)
		for(ind = 0; ind < rdeg[i]; ++ind)
		{
			for(j=0; j<St->rows; ++j)
				B[j] += Q[d2w[i]+ind][j] ;
		}

			
// Add the two				
//	printf("b\n");

	for(i=0; i<n; ++i)
		for(ind = 0; ind <rdeg[i];++ind)
			for(j=0; j<St->rows; ++j)
				Z[ d2w[i]+rdeg[i]+1+ind ][j] += lam*A[j];

//	printf("%d %d\n",cdeg[2],cdeg[1]);

	for(i=n; i<m+n; ++i)
		for(ind = 0; ind <cdeg[i-n];++ind)
			for(j=0; j<St->rows; ++j)
				{
					Z[ d2w[i]+cdeg[i-n]+1+ ind ][j] += lam*B[j];
//					printf("%d %d %d\n",i,j,ind);
				}	
				
//	printf("c\n");

	free(B);
//	printf("b %d\n\n\n\n",St->rows);
	free(A);
}










void initW(int flag)
{
	int i,j;
	double sum1;

	for(i=0; i<n+m; ++i)
		for(j=0; j<n+m; ++j)
			sum1 += E[i][j] ;

	
	SMat ES;
	int count = 0;
	int *dv;
	dv = calloc(m+n,sizeof(int));
	ES = svdNewSMat(n+m,n+m,(int)(sum1)) ;

	for(i=0; i<n+m; ++i)
		for(j=0; j<n+m; ++j)
			if( (int) E[i][j] == 1 )
			{
				ES->rowind[count] = j;
				ES->value[count] = 1.0;
				dv[i] += 1;
				++count;
			}

	ES->rows = n+m;
	ES->cols = n+m;
	ES->vals = count ;	
	ES->pointr[0] = 0 ;

    for(i=1; i<= ES->cols; ++i)
    	ES->pointr[i] = ES->pointr[i-1] + dv[i-1] ;

	free(dv);

	
	double **SS;
	int sum = 0;
	double **E2;
	E2 = callocArray(n+m,n+m) ;
	SS = callocArray(n+m,n+m);

//	matrixmul(E,E,n+m,n+m,n+m,E21,0,1.0,0.0);
//	sparsemul(ES,E,n+m,n+m,n+m,E2,0) ;
	
	


	addeye(E2,n+m,W,1);


//if( flag == 1 )
{

	for(i=0; i<kgir; ++i)
	{
//		matrixmul(W,E2, n+m,n+m,n+m,SS,0,1.0,0.0);
		sparsemul(ES,W,n+m,n+m,n+m,E2,0) ;
		sparsemul(ES,E2,n+m,n+m,n+m,SS,0) ;

		addeye(SS,n+m,W,1);
	}
//	matrixmul(W,E,n+m,n+m,n+m,SS,0,1.0,0.0);
	sparsemul(ES,W,n+m,n+m,n+m,SS,0) ;

	addeye(SS,n+m,W,1);


}

	for(i=0; i<n+m; ++i)
		for(j=0; j<n+m; ++j)
			W[i][j] = max1( bipConst[i][j] - (double)(sign1( (int) W[i][j] )),0.0);

	sum1 = 0;
	for(i=0; i<m+n; ++i)
		for(j=0; j<m+n; ++j)
			sum1 += W[i][j] ;
	printf("%d ",(int)sum1);
//	fprintf(fout,"%d ",(int)sum1);


	for(i=n; i<m+n; ++i)
		for(j=0; j<n; ++j)
			W[i][j] *= (double) (rdeg[j]*cdeg[i-n] ) ;
	
	for(i=0; i<n; ++i)
		for(j=n; j<m+n; ++j)
			W[i][j] *= (double) (rdeg[i]*cdeg[j-n] ) ;
	sum1 = 0;
	for(i=0; i<m+n; ++i)
		for(j=0; j<m+n; ++j)
			sum1 += W[i][j] ;
	
	printf("%d ",(int)sum1);
//	fprintf(fout,"%d ",(int)sum1);

	freeArray(E2,n+m);
	freeArray(SS,n+m);
	svdFreeSMat(ES);

}


int addedge(void) 
{
	double qsum = 0.0;
	int i,j;
	double *rowsum;
	rowsum = calloc(n+m,sizeof(double));

	for(i=0; i<n+m; ++i)
		for(j=0; j<n+m; ++j)
		{
			qsum += W[i][j] ;
			rowsum[i] += W[i][j];
		}

	
	printf("%e ",qsum) ;
//	fprintf(fout,"%e\n",qsum) ;
	if(qsum < 1e-300)
	{
		printf("\n");
		return(0);
	}	



	i = pick(rowsum,n+m);
	j = pick(W[i],n+m);
	int ri, cj;

	if(i < n )
	{
		ri = i;
		cj = j-n;
	}
	else
	{
		ri = j;
		cj = i-n;
	}

	rdeg[ri] -= 1;
	cdeg[cj] -= 1;
	E[ri][n+cj] = 1;
	E[n+cj][ri] = 1;

	printf("%d %d %d %d\n",rdeg[ri],cdeg[cj],ri,cj);
//	printf("%d %d\n",ri,cj);

	free(rowsum);
	return(1) ;
}	


int pick(double *probs,int nmax)
{
	int i;
	double sum = 0;
	for(i=0; i<nmax; ++i)
		sum += probs[i] ;
	
	double r;
	r = rannyu() * sum ;

	double cumsum = probs[0] ;
	for(i=0; i<nmax; ++i)
		if( r <= cumsum )
			return(i);
		else
			cumsum += probs[i+1];

}



void output(char* outfile,int t)
{
	int i,j;
	FILE *fout;
	char str[100] = "BipS.n";
	char str1[50];


	sprintf(str1,"%d",n);
	strcat(str,str1);
	strcat(str,".l");
	
	sprintf(str1,"%d",deg1) ;
	strcat(str,str1);
	strcat(str,".m");
	
	sprintf(str1,"%d",m) ;
	strcat(str,str1);
	strcat(str,".r");
	
	sprintf(str1,"%d",deg2) ;
	strcat(str,str1);
	strcat(str,".k");

	sprintf(str1,"%d",kgir) ;
	strcat(str,str1);
/*	strcat(str,".e");
	
	sprintf(str1,"%d",t) ;
	strcat(str,str1);
*/	strcat(str,".");
	strcat(str,outfile);


	fout = fopen(str,"w");

	fprintf(fout,"%d\n",n);
	fprintf(fout,"%d\n",m);

	for(i=n; i<n+m; ++i)
		for(j=0; j<n; ++j)
			if(E[i][j] == 1)
				fprintf(fout,"%d %d\n",j,i-n);
				

	fclose(fout);



}




void addeye(double** SS, int n1,double **out,int flag)
{
	int i,j;

	if( flag == 1 )
		for(i=0; i<n1; ++i)
			for(j=0; j<n1; ++j)
				out[i][j] = SS[i][j] ;
	

	for(i=0; i<n1; ++i)
		out[i][i] = SS[i][i] + 1;
		
}



void alloc_mem(void)
{
	int i,j;

	rdeg = calloc(n,sizeof(int));
	cdeg = calloc(m,sizeof(int));

	W  = callocArray(n+m,n+m);
	Lam  = callocArray(n+m,n+m);

	bipConst  = callocArray(n+m,n+m);
	E         = callocArray(n+m,n+m);
	
	for(i=0; i<n+m; ++i)
		E[i][i] = 1;

	for(i=n; i<n+m; ++i)
		for(j=0; j<n; ++j)
			bipConst[i][j] = 1;
	
	for(i=0; i<n; ++i)
		for(j=n; j<n+m; ++j)
			bipConst[i][j] = 1;
			

	d2w = calloc(n+m+1,sizeof(int));
	

}

void free_mem(void)
{
	int i;

	free(rdeg);
	free(cdeg);

	freeArray(Lam,n+m);
	freeArray(W,n+m);
	freeArray(E,n+m);
	freeArray(bipConst,n+m);

	free(d2w);
}



double factorial(int a, int b)
{
/****************************************
 * Compute a!/b! with the convention
 * that x! = 1 for x <= 0 
 * *************************************/

	if( a <= 0 )
		return( factorial(1,b) ) ;
	
	if( b<=0 )
		return( factorial(a,1) ) ;
	
	if( a == b)
		return(1.0) ;

	if( a > b )
		return( (a)*factorial(a-1,b) );
	
	if( a < b)
		return( factorial(a,b-1) / b ) ;
	
}	

double** callocArray(int n1, int n2)
{
        int i;
        double **A;
        A = calloc(n1,sizeof(double*));
        for(i=0; i<n1; ++i)
                A[i] = calloc(n2,sizeof(double));
        return(A) ;
}


double** mallocArray(int n1, int n2)
{
        int i;
        double **A;
        A = malloc(n1*sizeof(double*));
        for(i=0; i<n1; ++i)
                A[i] = malloc(n2*sizeof(double));
        return(A) ;
}

void freeArray(double **A, int n1)
{
        int i;
        for(i=0; i<n1; ++i)
                free(A[i]);
        free(A);
}


int sign1(int x)
{
	if( x == 0 )
		return(0) ;
	if( x > 0 ) 
		return(1);
	if( x < 0 )
		return(-1);

}

double max1(double a, double b)
{
	if( a > b )
		return(a);
	else
		return(b) ;

}

